"""
========================================
Calculations of specular differentiation
========================================

This module provides implementations of the function :math:`\\mathcal{A}`, specular directional derivatives, specular partial derivatives, specular derivatives, specular gradients, and specular Jacobians.

Computations are based on the finite difference approximation of one-sided (directional) derivatives.
"""

from typing import Callable
import math
import numpy as np


def A(
    alpha: float | np.number | int | list | np.ndarray,
    beta: float | np.number | int | list | np.ndarray,
    zero_tol: float = 1e-8
) -> float | np.ndarray:
    """
    Compute the function :math:`\\mathcal{A}` from one-sided directional derivatives.

    Given real numbers ``alpha`` and ``beta``, the function :math:`\\mathcal{A}:\\mathbb{R}^2 \\to \\mathbb{R}` is defined by 
    
        :math:`\\mathcal{A}(\\alpha, \\beta) = \\frac{\\alpha \\beta - 1 + \\sqrt((1 + \\alpha^2)(1 + \\beta^2))}{\\alpha + \\beta}` if ``alpha + beta != 0``; otherwise, it returns ``0``.

    Parameters
    ----------
    alpha : float | np.number | int | list | np.ndarray
        One-sided directional derivative.
    beta : float | np.number | int | list | np.ndarray
        One-sided directional derivative.
    zero_tol : float, optional
        A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.
        Default: ``1e-8``.

    Returns
    -------
    float
        The function :math:`\\mathcal{A}`.

    Raises
    ------
    ValueError
        If ``alpha`` and ``beta`` have different shape.

    Examples
    --------
    >>> import specular
    >>> specular.calculation.A(1.0, 2.0)
    1.3874258867227933
    >>> specular.calculation.A([1.0, 2.4], [2.0, 4.1])
    array([1.38742589, 3.04807583])
    """
    if np.isscalar(alpha) and np.isscalar(beta):
        return _A_scalar(alpha, beta, zero_tol=zero_tol) # type: ignore
    
    return _A_vector(alpha, beta, zero_tol=zero_tol) # type: ignore


def _A_scalar(
    alpha: float | np.number, 
    beta: float | np.number, 
    zero_tol: float = 1e-8
) -> float:
    """Scalar implementation of :func:`A`."""
    denominator = alpha + beta

    if abs(denominator) <= zero_tol:
        return 0.0
    
    numerator = alpha * beta - 1.0 + math.sqrt((1.0 + alpha**2) * (1.0 + beta**2))

    return numerator / denominator # type: ignore


def _A_vector(
    alpha: list | np.ndarray, 
    beta: list | np.ndarray, 
    zero_tol: float = 1e-8
) -> np.ndarray:
    """Vector implementation of :func:`A`."""
    alpha = np.asanyarray(alpha, dtype=float)
    beta = np.asanyarray(beta, dtype=float)

    if alpha.shape != beta.shape:
        raise ValueError(f"Shape mismatch: alpha {alpha.shape} vs beta {beta.shape}")

    denominator = alpha + beta

    mask = np.abs(denominator) > zero_tol
    result = np.zeros_like(denominator)

    alpha_valid = alpha[mask]
    beta_valid = beta[mask]
    denominator_valid = denominator[mask]

    if alpha_valid.size > 0:
        numerator = alpha_valid * beta_valid - 1.0 + np.sqrt((1.0 + alpha_valid**2) * (1.0 + beta_valid**2))
        
        result[mask] = numerator / denominator_valid

    return result


def derivative(
    f: Callable[[int | float | np.number], int | float | np.number | list | np.ndarray],
    x: float | np.number | int,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> float | np.ndarray:
    """
    Approximates the specular derivative of a function :math:`f:\\mathbb{R} \\to \\mathbb{R}^m` at a scalar point ``x``.
    
    If ``f`` returns a scalar, the result is a float.
    If ``f`` returns a vector, the result is a vector (component-wise derivative).


    Parameters
    ----------
    f : callable
        A function of a single real variable, returning a scalar or a vector.
    x : float | np.number | int 
        The point at which the derivative is evaluated.
    h : float, optional
        Mesh size used in the finite difference approximation. Must be positive.
        Default: ``1e-6``.
    zero_tol : float, optional 
        A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.
        Default: ``1e-8``.

    Returns
    -------
    float | np.ndarray
        The approximated specular derivative of ``f`` at ``x``.

    Raises
    ------
    TypeError
        If the type of ``x`` is not a scalar.
    ValueError
        If ``h`` is not positive.
    
    Examples
    --------
    >>> import specular
    >>> f = lambda x: max(x, 0.0)
    >>> specular.derivative(f, x=0.0)
    0.41421356237309515
    >>> f = lambda x: abs(x)
    >>> specular.derivative(f, x=0.0)
    0.0
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")
    
    try:
        x = float(x)

    except TypeError:
        raise TypeError(
            f"Input 'x' must be a scalar. "
            f"Got {type(x).__name__}. "
            "Use `specular.directional_derivative`, `specular.gradient`, or `specular.jacobian` for vectors inputs."
        )

    f_val = f(x)

    # f is real-valued
    if np.ndim(f_val) == 0:
        alpha = (f(x + h) - f_val) / h # type: ignore
        beta = (f_val - f(x - h)) / h # type: ignore
        
        return _A_scalar(alpha, beta, zero_tol=zero_tol) # type: ignore

    # f is vector-valued
    else:
        f_val = np.asarray(f_val, dtype=float)
        f_right = np.asarray(f(x + h), dtype=float)
        f_left = np.asarray(f(x - h), dtype=float)

        alpha = (f_right - f_val) / h
        beta = (f_val - f_left) / h
        
        return _A_vector(alpha, beta, zero_tol=zero_tol)


def directional_derivative(
    f: Callable[[list | np.ndarray], int | float | np.number],
    x: list | np.ndarray,
    v: list | np.ndarray,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> float:
    """
    Approximates the specular directional derivative of a function :math:`f:\\mathbb{R}^n \\to \\mathbb{R}` at a point ``x`` in the direction ``v``.
    The one-sided derivatives ``alpha`` and ``beta`` are approximated.
    The function :math:`\\mathcal{A}` returns the specular directional derivative for each component of the vector ``x``.

    Parameters
    ----------
    f : callable
        A real-valued function defined on an open subset of :math:`\\mathbb{R}^n`.
    x : list | np.ndarray
        The point at which the derivative is evaluated.
    v : list | np.ndarray
        The direction in which the derivative is taken.
    h : float, optional
        Mesh size used in the finite difference approximation. Must be positive.
        Default: ``1e-6``.
    zero_tol : float, optional
        A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.
        Default: ``1e-8``.

    Returns
    -------
    float
        The approximated specular directional derivative of ``f`` at ``x`` in the direction ``v`` as a scalar.

    Raises
    ------
    TypeError
        If ``x`` or ``v`` are not of valid array-like types.
    ValueError
        If ``x`` and ``v`` have different shape.
        If ``h`` is not positive.

    Examples
    --------
    >>> import specular
    >>> import math
    >>> f = lambda x: math.sqrt(x[0]**2 + x[1]**2 + x[2]**2)
    >>> specular.directional_derivative(f, x=[0.0, 0.1, -0.1], v=[1.0, -1.0, 2.0])
    -2.1213203434708223
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")
    
    x = np.asarray(x, dtype=float)
    v = np.asarray(v, dtype=float)

    if x.ndim == 0:
        raise TypeError(
            f"Input 'x' must be a vector. "
            f"Got {type(x).__name__}. "
            "Use `specular.derivative` for scalar inputs."
        )
    
    if v.ndim == 0:
        raise TypeError(
            "Input 'v' must be a vector. "
            f"Got {type(v).__name__}."
        )
    
    if x.shape != v.shape:
        raise ValueError(f"Shape mismatch: x {x.shape} vs v {v.shape}")
    
    f_val = f(x) 

    if np.ndim(f_val) != 0:
        raise ValueError(
            "Function f must return a scalar value. "
            f"Got shape {np.shape(f_val)}."
        )
    
    alpha = (f(x + h * v) - f_val)/h
    beta = (f_val - f(x - h * v))/h

    return _A_scalar(alpha, beta, zero_tol=zero_tol)


def partial_derivative(
    f: Callable[[list | np.ndarray], int |float | np.number],
    x: list | np.ndarray,
    i: int | np.integer,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> float:
    """
    Approximates the i-th specular partial derivative of a real-valued function :math:`f:\\mathbb{R}^n \\to \\mathbb{R}` at point ``x`` for ``n > 1``.

    This is computed using :func:`specular_directional_derivative` with the direction of the ``i``-th standard basis vector of :math:`\\mathbb{R}^n`.

    Parameters
    ----------
    f : callable
        A real-valued function defined on :math:`\\mathbb{R}^n`.
        ``n`` is the dimension of ``x``.
    x : list | np.ndarray
        The point at which the derivative is evaluated.
    i : int | np.integer
        The index of the specular partial derivative with respect to :math:`x_i` (``1 <= i <= n``).
    h : float, optional
        Mesh size used in the finite difference approximation. Must be positive.
        Default: ``1e-6``.
    zero_tol : float, optional
        A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.
        Default: ``1e-8``.

    Returns
    -------
    float
        The approximated ``i``-th partial specular derivative of ``f`` at ``x`` as a scalar.

    Raises
    ------
    TypeError
        If ``i`` is not an integer.
    ValueError
        If ``i`` is out of the valid range (``1 <= i <= n``).

    Examples
    --------
    >>> import specular
    >>> import math 
    >>> f = lambda x: math.sqrt(x[0]**2 + x[1]**2 + x[2]**2)
    >>> specular.partial_derivative(f, x=[0.1, 2.3, -1.2], i=2)
    0.8859268982863702
    """
    x = np.asarray(x, dtype=float)

    if not isinstance(i, (int, np.integer)):
        raise TypeError(f"Index 'i' must be an integer. Got {type(i).__name__}")

    n = x.size
    if i < 1 or i > n:
        raise ValueError(f"Index 'i' must be between 1 and {n} (dimension of x). Got {i}")

    e_i = np.zeros_like(x)
    e_i[i - 1] = 1.0

    return directional_derivative(f, x, e_i, h, zero_tol)


def gradient(
    f: Callable[[list | np.ndarray], int |float | np.number],
    x: list | np.ndarray,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> np.ndarray:
    """
    Approximates the specular gradient of a real-valued function :math:`f:\\mathbb{R}^n \\to \\mathbb{R}` at point ``x`` for ``n > 1``.

    The specular gradient is defined as the vector of all partial specular derivatives along the standard basis directions.

    Parameters
    ----------
    f : callable
        A real-valued function defined on :math:`\\mathbb{R}^n`.
    x : list | np.ndarray
        The point at which the specular gradient is evaluated.
    h : float, optional
        Mesh size used in the finite difference approximation. Must be positive.
        Default: ``1e-6``.
    zero_tol : float, optional
        A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.
        Default: ``1e-8``.

    Returns
    -------
    np.ndarray
        The approximated specular gradient of ``f`` at ``x`` as a vector.

    Raises
    ------
    TypeError
        If ``x`` is not of a valid array-like type.
    ValueError
        If ``f`` does not return a scalar value.
        If ``h`` is not positive.

    Examples
    --------
    >>> import specular
    >>> import numpy as np
    >>> f = lambda x: np.linalg.norm(x)
    >>> specular.gradient(f, x=[1.4, -3.47, 4.57, 9.9])
    array([ 0.12144298, -0.3010051 ,  0.39642458,  0.85877534])
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")
    
    x = np.asarray(x, dtype=float)
    
    if x.ndim != 1:
        raise TypeError(
            f"Input 'x' must be a vector. "
            f"Got {type(x).__name__} with shape {x.shape}. "
            "Use `specular.derivative` for scalar inputs."
        )
    
    n = x.size
    I = np.eye(n)

    f_val = f(x)

    if np.ndim(f_val) != 0:
        raise ValueError(
            "Function f must return a scalar value. "
            f"Got shape {np.shape(f_val)}."
        )

    x_right = x + h*I
    x_left = x - h*I

    f_right = np.array([f(row) for row in x_right], dtype=float)
    f_left = np.array([f(row) for row in x_left], dtype=float)
    
    # try:
    #     f_right = f(x_right)
    #     if np.ndim(f_right) != 1 or np.size(f_right) != n:
    #         raise ValueError (
    #             "Function f must return a scalar for each input vector. "
    #             f"Got shape {np.ndim(f_right)}."
    #         )
        
    # except Exception:
    #     f_right = np.array([f(row) for row in x_right])

    # try:
    #     f_left = f(x_left)
    #     if np.ndim(f_left) != 1 or np.size(f_left) != n:
    #         raise ValueError(
    #             "Function f must return a scalar for each input vector. "
    #             f"Got shape {np.ndim(f_left)}.")
        
    # except Exception:
    #     f_left = np.array([f(row) for row in x_left])

    alpha = (f_right - f_val) / h 
    beta = (f_val - f_left) / h 

    return _A_vector(alpha, beta, zero_tol=zero_tol) # type: ignore


def jacobian(
    f: Callable[[list | np.ndarray], int | float | np.number | list | np.ndarray],
    x: list | np.ndarray,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> np.ndarray:
    """
    Approximates the specular Jacobian matrix of a vector-valued function :math:`f:\\mathbb{R}^n \\to \\mathbb{R}^m`.

    Returns a matrix of shape (m, n) where J[j, i] is the partial derivative of f_j with respect to x_i (1 <= i <= n, 1 <= j <= m).

    Raises
    ------
    TypeError
        If ``x`` is not of a valid array-like type.
    ValueError
        If ``h`` is not positive.
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")

    x = np.asarray(x, dtype=float)

    if x.ndim != 1:
        raise TypeError(
            f"Input 'x' must be a vector. "
            f"Got {type(x).__name__} with shape {x.shape}. "
            "Use `specular.derivative` for scalar inputs."
        )

    n = x.size

    f_val = np.asarray(f(x), dtype=float)

    if f_val.ndim == 0:
        f_val = f_val.reshape(1)

    m = f_val.size

    I = np.eye(n)
    
    x_right = x + h * I
    x_left = x - h * I

    # try:
    #     f_right = np.asarray(f(x_right), dtype=float)
    #     f_left = np.asarray(f(x_left), dtype=float)

    #     if f_right.size != n * m or f_left.size != n * m:
    #         raise ValueError("Vectorization mismatch")

    # except Exception:
    #     f_right = np.array([f(row) for row in x_right], dtype=float)
    #     f_left = np.array([f(row) for row in x_left], dtype=float)

    # f_right = np.asarray(f_right, dtype=float).reshape(n, m)
    # f_left = np.asarray(f_left, dtype=float).reshape(n, m)

    f_right = np.array([f(row) for row in x_right], dtype=float)
    f_left = np.array([f(row) for row in x_left], dtype=float)

    f_right = f_right.reshape(n, m)
    f_left = f_left.reshape(n, m)

    alpha = (f_right - f_val) / h
    beta = (f_val - f_left) / h

    J_transposed = _A_vector(alpha, beta, zero_tol=zero_tol)

    return J_transposed.T


from .result import ODEResult
from .solver import (
    classical_scheme,
    Euler_scheme,
    trigonometric_scheme
)

__all__ = [
    "classical_scheme",
    "Euler_scheme",
    "trigonometric_scheme",
    "ODEResult"
]


import math
import numpy as np
from typing import Callable, Tuple

class StepSize:
    """
    Step size rules for optimization methods.
    """
    __options__ = [
        'constant',
        'not_summable',
        'square_summable_not_summable',
        'geometric_series',
        'user_defined'
    ]

    def __init__(
        self,
        name: str,
        parameters: float | np.floating | int | Tuple | list | np.ndarray | Callable
    ):
        """
        The step size rules for optimization methods:

        :math:`x_{k+1} = x_k - h_k s_k`,
        
        where :math:`s_k` is the search direction and :math:`h_k > 0` is the step size at iteration `k >= 1`.

        Parameters
        ----------
        name : str
            Options: 'constant', 'not_summable', 'square_summable_not_summable', 'geometric_series', 'user_defined'
        parameters : float | int | tuple | list | np.ndarray | Callable
            The parameters required for the selected step size rule:

            * 'constant': float or int

                A number `a > 0` for the rule `h_k = a` for each `k`.
            
            * 'not_summable': float or int

                A number `a > 0` for the rule `h_k = a / sqrt{k}` for each `k`.
            
            * 'square_summable_not_summable': list or tuple

                A pair of numbers `[a, b]`, where `a > 0` and `b >= 0`, for the rule `a / (b + k)` for each `k`.
            
            * 'geometric_series': list or tuple

                A pair of numbers `[a, r]`, where `a > 0` and `0 < r < 1`, for the rule :math:`a * r^k` for each `k`.
            
            * 'user_defined': Callable

                A function that takes the current iteration `k` as input and returns the step size (float).
        """
        self.step_size = name
        self.parameters = parameters

        init_methods = {
            'constant': self._init_constant,
            'not_summable': self._init_not_summable,
            'square_summable_not_summable': self._init_square_summable,
            'geometric_series': self._init_geometric,
            'user_defined': self._init_user_defined
        }

        if name not in init_methods:
             raise ValueError(f"Invalid step size '{name}'. Options: {self.__options__}")
        
        init_methods[name]()

    def __call__(self, k: int) -> float:
        """
        Returns the step size at iteration k.
        """
        return self._rule(k)

    # ==== Initialization Methods ====
    def _init_constant(self):
        if not isinstance(self.parameters, (float, int, np.floating)):
            raise TypeError(f"Invalid type: number required. Got {type(self.parameters)}")

        if self.parameters <= 0:
            raise ValueError(f"Invalid value: positive number required. Got {self.parameters}")

        self.a = float(self.parameters)
        self._rule = self._calc_constant

    def _init_not_summable(self):
        if not isinstance(self.parameters, (float, int, np.floating)):
            raise TypeError(f"Invalid type: number required. Got {type(self.parameters)}")

        if self.parameters <= 0:
            raise ValueError(f"Invalid value: positive number required. Got {self.parameters}")

        self.a = float(self.parameters)
        self._rule = self._calc_not_summable

    def _init_square_summable(self):
        if not isinstance(self.parameters, (tuple, list, np.ndarray)):
            raise TypeError(f"Invalid type: list/tuple required. Got {type(self.parameters)}")

        if len(self.parameters) != 2:
            raise ValueError(f"Invalid length: 2 parameters [a, b] required. Got {len(self.parameters)}")

        self.a, self.b = self.parameters[0], self.parameters[1]

        if self.a <= 0 or self.b < 0:
            raise ValueError(f"Invalid parameters: a > 0 and b >= 0 required. Got a={self.a}, b={self.b}")

        self._rule = self._calc_square_summable_not_summable

    def _init_geometric(self):
        if not isinstance(self.parameters, (tuple, list, np.ndarray)):
            raise TypeError(f"Invalid type: list/tuple required. Got {type(self.parameters)}")
        
        if len(self.parameters) != 2:
            raise ValueError(f"Invalid length: 2 parameters [a, r] required. Got {len(self.parameters)}")

        self.a, self.r = self.parameters[0], self.parameters[1]

        if self.a <= 0 or not (0.0 < self.r < 1.0):
            raise ValueError(f"Invalid parameters: a > 0 and 0 < r < 1 required. Got a={self.a}, r={self.r}")

        self._rule = self._calc_geometric_series

    def _init_user_defined(self):
        if not callable(self.parameters):
            raise TypeError("Invalid type: callable function required.")

        self._rule = self.parameters

    # ==== Calculation Methods ====
    def _calc_constant(self, k: int) -> float:
        """
        h_k = a 
        """
        return self.a

    def _calc_not_summable(self, k: int) -> float:
        """
        h_k = a / sqrt{k}
        """
        return self.a / math.sqrt(k)

    def _calc_square_summable_not_summable(self, k: int) -> float:
        """
        h_k = a / (b + k)
        """
        return self.a / (self.b + k)

    def _calc_geometric_series(self, k: int) -> float:
        """
        h_k = a * r**k
        """
        return self.a * (self.r ** k)


from .calculation import (
    A,
    derivative,
    directional_derivative,
    partial_derivative,
    gradient,
    jacobian
)

__all__ = [
    "A",
    "derivative",
    "directional_derivative",
    "partial_derivative",
    "gradient",
    "jacobian",
]


from .result import OptimizationResult
from .step_size import StepSize
import time
import torch
import numpy as np
from scipy.optimize import minimize
from typing import Callable, Union

def gradient_descent_method(
    f_torch: Callable[[torch.Tensor], torch.Tensor], 
    x_0: Union[np.ndarray, list], 
    step_size: StepSize, 
    max_iter: int = 100
) -> OptimizationResult:
    """
    Performs optimization using standard gradient descent.

    Returns
    -------
    OptimizationResult
        The result of the optimization containing the solution, function value, number of iterations, runtime, and history.
    """
    start_time = time.time()
    
    x = torch.tensor(x_0, dtype=torch.float32, requires_grad=True)

    x_history = [x.detach().cpu().numpy().copy()]
    f_history = [f_torch(x.detach()).item()]

    for k in range(1, max_iter + 1):
        if x.grad is not None:
            x.grad.zero_()

        loss = f_torch(x)
        loss.backward()

        with torch.no_grad():
            if x.grad is not None:
                x -= step_size(k) * x.grad
        
        x_history.append(x.detach().cpu().numpy().copy())
        f_history.append(loss.item())

    end_time = time.time()

    return OptimizationResult(
        method="gradient descent",
        solution=x_history[-1],
        func_val=f_history[-1],
        iteration=max_iter,
        runtime=end_time - start_time,
        all_history={
            "variables": np.array(x_history),
            "values": np.array(f_history)
        }
    )

def Adam(
    f_torch: Callable[[torch.Tensor], torch.Tensor],
    x_0: Union[np.ndarray, list],
    step_size: StepSize | float,
    max_iter: int = 100
) -> OptimizationResult:
    """
    Performs optimization using the Adam algorithm from PyTorch.
    
    Returns
    -------
    OptimizationResult
        The result of the optimization containing the solution, function value, number of iterations, runtime, and history.
    """
    start_time = time.time()
    
    x = torch.tensor(x_0, dtype=torch.float32, requires_grad=True)
    
    initial_lr = step_size(1) if callable(step_size) else step_size
    optimizer = torch.optim.Adam([x], lr=initial_lr)

    x_history = [x.detach().cpu().numpy().copy()]
    f_history = [f_torch(x).item()]

    for k in range(1, max_iter + 1):
        if callable(step_size):
            current_lr = step_size(k)

            for param_group in optimizer.param_groups:
                param_group['lr'] = current_lr
        
        optimizer.zero_grad()
        loss = f_torch(x)
        loss.backward()
        optimizer.step()
        
        x_history.append(x.detach().cpu().numpy().copy())
        f_history.append(loss.item())

    end_time = time.time()
    
    return OptimizationResult(
        method="Adam",
        solution=x_history[-1],
        func_val=f_history[-1],
        iteration=max_iter,
        runtime=end_time - start_time,
        all_history={
            "variables": np.array(x_history),
            "values": np.array(f_history)
        }
    )

def BFGS(
    f_np: Callable[[np.ndarray], float], 
    x_0: np.ndarray, 
    max_iter: int = 100, 
    tol: float = 1e-6
) -> OptimizationResult:
    """
    Performs optimization using the BFGS algorithm from SciPy.

    Returns
    -------
    OptimizationResult
        The result of the optimization containing the solution, function value, number of iterations, runtime, and history.
    """
    start_time = time.time()
    
    x_history = [np.array(x_0).copy()]
    f_history = [f_np(x_0)]

    def bfgs_callback(x_k):
        x_val = np.array(x_k).copy()
        f_val = f_np(x_k)
        
        x_history.append(x_val)
        f_history.append(f_val)

    result = minimize(
        f_np,
        x_0,
        method='BFGS',
        callback=bfgs_callback,
        options={'maxiter': max_iter, 'gtol': tol}
    )

    end_time = time.time()

    return OptimizationResult(
        method="BFGS",
        solution=result.x,
        func_val=result.fun, 
        iteration=result.nit,
        runtime=end_time - start_time,
        all_history={
            "variables": np.array(x_history),
            "values": np.array(f_history)
        }
    )



import pytest
import math
import numpy as np
import specular

# ==========================================
# 1. Test Function A (Core Logic)
# ==========================================
def test_A_scalar():
    """Test scalar inputs for A."""
    expected = 1.0
    assert specular.A(1.0, 1.0) == pytest.approx(expected)
    
    expected = math.sqrt(2) - 1
    assert specular.A(0.0, 1.0) == pytest.approx(expected)

def test_A_vector():
    """Test vector inputs for A (using numpy arrays)."""
    alpha = np.array([1.0, 0.0])
    beta = np.array([1.0, 1.0])
    
    result = specular.A(alpha, beta)
    
    expected_0 = 1.0
    expected_1 = math.sqrt(2) - 1
    
    assert result[0] == pytest.approx(expected_0) # type: ignore
    assert result[1] == pytest.approx(expected_1) # type: ignore

# ==========================================
# 2. Test Derivative (Scalar Input)
# ==========================================
def test_derivative_scalar_output():
    """f: R -> R (Smooth function: x^2)."""
    # f(x) = x**2
    f = lambda x: x**2
    assert specular.derivative(f, x=2.0) == pytest.approx(4.0, rel=1e-4)

def test_derivative_specular_property():
    """f: R -> R (Nonsmooth function: |x|)."""
    # f(x) = |x|
    f = lambda x: abs(x)
    assert specular.derivative(f, x=0.0) == pytest.approx(0.0, abs=1e-6)

def test_derivative_vector_output():
    """f: R -> R^2 (Parametric curve)."""
    # f(x) = (x, x**2)
    f = lambda x: [x, x**2]
    
    result = specular.derivative(f, x=2.0)
    
    np.testing.assert_allclose(result, [1.0, 4.0], rtol=1e-4)

# ==========================================
# 3. Test Directional Derivative
# ==========================================
def test_directional_derivative():
    """f: R^2 -> R."""
    # f(x_1, x_2) = x_1**2 + x_2**2
    f = lambda v: v[0]**2 + v[1]**2
    x = [1.0, 1.0]
    
    assert specular.directional_derivative(f, x, v=[1.0, 0.0]) == pytest.approx(2.0, rel=1e-4)

# ==========================================
# 4. Test Partial Derivative
# ==========================================
def test_partial_derivative():
    """Test partial derivative with 1-based indexing."""
    # f(x_1, x_2, x_3) = x_1 + 2x_2 + 3x_3
    f = lambda x: x[0] + 2*x[1] + 3*x[2]
    x = [0, 0, 0]
    
    assert specular.partial_derivative(f, x, i=2) == pytest.approx(2.0, rel=1e-4)
    assert specular.partial_derivative(f, x, i=3) == pytest.approx(3.0, rel=1e-4)

def test_partial_derivative_error():
    """Check index out of bounds error."""
    # f(x_1, x_2) = x_1
    f = lambda x: x[0]
    x = [1.0, 2.0]

    with pytest.raises(ValueError):
        specular.partial_derivative(f, x, i=3)

# ==========================================
# 5. Test Gradient
# ==========================================
def test_gradient():
    """f: R^3 -> R."""
    # f(x_1, x_2, x_3) = x_1 + x_2 + x_3
    f = lambda x: np.sum(np.square(x))
    x = [1.0, 2.0, 3.0]
    
    grad = specular.gradient(f, x)
    
    np.testing.assert_allclose(grad, [2.0, 4.0, 6.0], rtol=1e-4)

# ==========================================
# 6. Test Jacobian
# ==========================================
def test_jacobian():
    """f: R^2 -> R^2."""
    # f(x_1, x_2) = (x**2, x_1 + x_2)
    f = lambda x: [x[0]**2, x[0] + x[1]]
    x = [2.0, 1.0]
    
    J = specular.jacobian(f, x)
    
    expected_J = np.array([
        [4.0, 0.0],
        [1.0, 1.0]
    ])
    
    assert J.shape == (2, 2)
    np.testing.assert_allclose(J, expected_J, rtol=1e-4)

def test_jacobian_scalar_output():
    """Check if Jacobian works for scalar output (should be 1xN matrix)."""
    # f(x_1, x_2, x_3) = x_1 + x_2 + x_3
    f = lambda x: np.sum(x)
    x = [1.0, 2.0, 3.0]
    
    J = specular.jacobian(f, x)
    assert J.shape == (1, 3)
    np.testing.assert_allclose(J, [[1.0, 1.0, 1.0]], rtol=1e-4)


"""
==============================================================
Numerical methods for solving ordinary differential equations
==============================================================

Let the source function F:[t_0, T]xR -> R be given, and the initial data u_0:R -> R be given. 
Consider the initial value problem:
(IVP)              u'(t) = F(t, u(t))
with the initial condition u(t_0) = u_0(t_0).
To solve (IVP) numerically, this module provides implementations of the specular Euler schemes, the Crank-Nicolson scheme, and the specular trigonometric scheme.
"""

import math
import numpy as np
from tqdm import tqdm 
from typing import Optional, Callable, Tuple, List
from .result import ODEResult
from ..calculation import _A_scalar as A

SUPPORTED_SCHEMES = ['explicit Euler', 'implicit Euler', 'Crank-Nicolson']

def classical_scheme(
    F: Callable[[float, float], float], 
    t_0: float, 
    u_0: Callable[[float], float] | float,
    T: float, 
    h: float = 1e-6,
    form: str = 'explicit Euler',
    tol: float = 1e-6, 
    max_iter: int = 100
) -> ODEResult:
    """
    Solves an initial value problem (IVP) using classical numerical schemes.
    Supported forms: explicit Euler, implicit Euler, and Crank-Nicolson.

    Parameters
    ----------
    F : callable
        The given source function ``F`` in (IVP).
        The calling signature should be ``F(t, u)`` where ``t`` and ``u`` are scalars.
    t_0 : float
        The starting time of the simulation.
    u_0 : callable
        The given initial condition ``u_0`` in (IVP).
    T : float
        The end time of the simulation.
    h : float, optional
        Mesh size used in the finite difference approximation. Must be positive.
        Default: ``1e-6``.
    form : str | optional
        The form of the numerical scheme. 
        Options: ``'explicit_Euler'``, ``'implicit_Euler'``, ``'Crank-Nicolson'``.
        Default: ``'explicit_Euler'``.
    tol : float | optional
        Tolerance for fixed-point iteration.
        Used for implicit Euler and Crank-Nicolson schemes.
    max_iter : int | optional
        Max iterations for fixed-point solver.

    Returns
    -------
    ODEResult
        An object containing ``(t, u)`` data and the scheme name.
    """
    t_curr = t_0
    u_curr = u_0(t_0) if callable(u_0) else u_0 

    all_history = {}
    t_history = [t_curr]
    u_history = [u_curr]

    steps = int((T - t_0) / h)
    
    if form == "explicit Euler":
        for _ in tqdm(range(steps), desc="Running the explicit Euler scheme"):
            t_curr, u_curr = t_curr + h, u_curr + h*F(t_curr, u_curr) # type: ignore

            t_history.append(t_curr) 
            u_history.append(u_curr) 

    elif form == "implicit Euler":
        for k in tqdm(range(steps), desc="Running the implicit Euler scheme"):
            t_next = t_curr + h

            # Initial guess: explicit Euler 
            u_temp = u_curr + h*F(t_curr, u_curr) # type: ignore
            u_guess = u_temp

            # Fixed-point iteration
            for _ in range(max_iter):
                u_guess = u_curr + h*F(t_next, u_temp)
                if np.linalg.norm(u_guess - u_temp) < tol:
                    break
                u_temp = u_guess
            else:
                print(f"Warning: step {k+1} did not converge.")

            t_curr, u_curr = t_next, u_guess  
            t_history.append(t_curr)
            u_history.append(u_curr)
    
    elif form == "Crank-Nicolson":
        for k in tqdm(range(steps), desc="Running Crank-Nicolson scheme"):
            t_next = t_curr + h

            F_curr = F(t_curr, u_curr)

            # Initial guess: explicit Euler
            u_temp = u_curr + h * F_curr
            u_guess = u_temp

            # Fixed-point iteration
            for _ in range(max_iter):
                f_guess = F(t_next, u_temp)
                u_guess = u_curr + 0.5 * h * (F_curr + f_guess)

                if np.linalg.norm(u_guess - u_temp) < tol:
                    break

                u_temp = u_guess
            else:
                print(f"Warning: step {k+1} did not converge.")

            t_curr, u_curr = t_next, u_guess
            t_history.append(t_curr)
            u_history.append(u_curr)
            
    else:
        raise ValueError(f"Unknown form '{form}'. Supported forms: {SUPPORTED_SCHEMES}")
    
    all_history["variables"] = np.array(t_history)
    all_history["values"] = np.array(u_history)

    return ODEResult(
        scheme= form + " scheme",
        h=h,
        all_history=all_history
    )

def Euler_scheme(
    of_Type: int | str,
    F: Callable[[float, float], float],
    t_0: float, 
    u_0: Callable[[float], float] | float,
    T: float, 
    h: float = 1e-6,
    u_1: Callable[[float], float] | float | bool = False,
    tol: float = 1e-6, 
    zero_tol: float = 1e-8,
    max_iter: int = 100
) -> ODEResult:
    """
    Solves an initial value problem (IVP) using the specular Euler scheme of Type 1, 2, 3, 4, 5, and 6.

    Parameters
    ----------
    of_Type : int | str
        The type of the specular Euler scheme.
        Options: ``1``, ``'1'``, ``2``, ``'2'``, ``3``, ``'3'``, ``4``, ``'4'``, ``5``, ``'5'``, ``6``, ``'6'``.
    F : callable
        The given source function ``F`` in (IVP).
        The calling signature should be ``F(t, u)`` where ``t`` and ``u`` are scalars.
    t_0 : float
        The starting time of the simulation.
    u_0 : callable
        The given initial condition ``u_0`` in (IVP).
    T : float
        The end time of the simulation.
    h : float, optional
        Mesh size used in the finite difference approximation. Must be positive.
        Default: ``1e-6``.
    u_1 : callable | float | bool
        The numerical solution at the time ``t_1 = t_0 + h`` for Types 1, 2, and 3.
        If a float or callable is provided, it is used as the exact value.
        If False, the explicit Euler scheme is applied.
        Default: ``False``.
    tol : float | optional
        Tolerance for fixed-point iteration
        Used for Types 3, 4, 5, and 6.
    zero_tol : float | np.floating
        A small threshold used to determine if the denominator (alpha + beta) is close to zero for numerical stability.
        Default: ``1e-6``.
    max_iter : int | optional
        Max iterations for fixed-point solver.

    Returns
    -------
    ODEResult
        An object containing ``(t, u)`` data and the scheme name.
    """
    Type = str(of_Type)

    scheme = 'specular Euler scheme of Type ' + Type
    steps = int((T - t_0) / h)

    all_history = {}
    t_history = []
    u_history = []

    if Type in ['1', '2', '3']:
        t_prev = t_0
        u_prev = u_0(t_0) if callable(u_0) else u_0

        t_history.append(t_prev)
        u_history.append(u_prev)

        t_curr = t_prev + h

        if u_1 == False:
            # explicit Euler to get u_1
            u_curr = u_prev + h * F(t_prev, u_prev)
        else:
            u_curr = u_1
        
        t_history.append(t_curr)
        u_history.append(u_curr)

        if Type == '1':
            _of_Type_1(F, h, zero_tol, steps, t_prev, t_curr, u_prev, u_curr, t_history, u_history)

        elif Type == '2':
            _of_Type_2(F, h, zero_tol, steps, t_prev, t_curr, u_prev, u_curr, t_history, u_history)

        elif Type == '3':
            _of_Type_3(F, h, tol, zero_tol, max_iter, steps, t_prev, t_curr, u_prev, u_curr, t_history, u_history)

    elif Type in ['4', '5', '6']:
        t_curr = t_0
        u_curr = u_0(t_0) if callable(u_0) else u_0  

        t_history.append(t_curr)
        u_history.append(u_curr)

        if Type == '4':
            _of_Type_4(F, h, tol, zero_tol, max_iter, steps, t_curr, u_curr, t_history, u_history)

        elif Type == '5':
            _of_Type_5(F, h, tol, zero_tol, max_iter, steps, t_curr, u_curr, t_history, u_history)

        elif Type == '6':
            _of_Type_6(F, h, tol, zero_tol, max_iter, steps, t_curr, u_curr, t_history, u_history)

    else:
        raise ValueError(f"Unknown type. Got {of_Type}. Supported types: '1', '2', '3', '4', '5', and '6'")

    all_history["variables"] = np.array(t_history)
    all_history["values"] = np.array(u_history)

    return ODEResult(
        scheme=scheme,
        h=h,
        all_history=all_history
    )

def _of_Type_1(F, h, zero_tol, steps, t_prev, t_curr, u_prev, u_curr, t_history, u_history):
    """
    Implements the specular Euler scheme of Type 1.
    """
    for _ in tqdm(range(steps - 1), desc="Running the specular Euler scheme of Type 1"):
        t_next = t_curr + h
        u_next = u_curr + h * A(F(t_curr, u_curr), F(t_prev, u_prev), zero_tol=zero_tol)

        # Update for next step
        t_prev, u_prev = t_curr, u_curr
        t_curr, u_curr = t_next, u_next

        t_history.append(t_curr)
        u_history.append(u_curr)

def _of_Type_2(F, h, zero_tol, steps, t_prev, t_curr, u_prev, u_curr, t_history, u_history):
    """
    Implements the specular Euler scheme of Type 2.
    """
    for _ in tqdm(range(steps - 1), desc="Running the specular Euler scheme of Type 2"):
        t_next = t_curr + h
        u_next = u_curr + h * A(F(t_curr, u_curr), (u_curr - u_prev)/h, zero_tol=zero_tol)

        # Update for next step
        t_prev, u_prev = t_curr, u_curr
        t_curr, u_curr = t_next, u_next

        t_history.append(t_curr)
        u_history.append(u_curr)

def _of_Type_3(F, h, tol, zero_tol, max_iter, steps, t_prev, t_curr, u_prev, u_curr, t_history, u_history):
    """
    Implements the specular Euler scheme of Type 3.
    """
    for k in tqdm(range(steps - 1), desc="Running the specular Euler scheme of Type 3"):
        t_next = t_curr + h

        # Initial guess: explicit Euler
        u_temp = u_curr + h * F(t_curr, u_curr) 
        u_guess = u_temp

        # fixed second argument
        beta = F(t_prev, u_prev)
        
        # Fixed-point iteration
        for _ in range(max_iter):
            alpha = (u_temp - u_curr) / h
            u_guess = u_curr + h * A(alpha, beta, zero_tol=zero_tol)

            if abs(u_guess - u_temp) < tol:
                break

            u_temp = u_guess
        else:
            print(f"Warning: fixed-point iteration did not converge at step {k+1}")

        # Update for next step
        t_prev, u_prev = t_curr, u_curr
        t_curr, u_curr = t_next, u_guess 

        t_history.append(t_curr)
        u_history.append(u_curr)

def _of_Type_4(F, h, tol, zero_tol, max_iter, steps, t_curr, u_curr, t_history, u_history):
    """
    Implements the specular Euler scheme of Type 4.
    """
    for k in tqdm(range(steps), desc="Running the specular Euler scheme of Type 4"):
        t_next = t_curr + h

        # Initial guess: explicit Euler
        u_temp = u_curr + h * F(t_curr, u_curr)
        u_guess = u_temp
        
        beta = F(t_curr, u_curr)  # fixed second argument

        # Fixed-point iteration
        for _ in range(max_iter):
            alpha = (u_temp - u_curr) / h
            u_guess = u_curr + h * A(alpha, beta, zero_tol=zero_tol) 

            if abs(u_guess - u_temp) < tol:
                break

            u_temp = u_guess
        else:
            print(f"Warning: fixed-point iteration did not converge at step {k+1}")

        # Update for next step
        t_curr, u_curr = t_next, u_guess  

        t_history.append(t_curr)
        u_history.append(u_curr)

def _of_Type_5(F, h, tol, zero_tol, max_iter, steps, t_curr, u_curr, t_history, u_history):
    """
    Implements the specular Euler scheme of Type 5.
    """
    for k in tqdm(range(steps), desc="Running the specular Euler scheme of Type 5"):
        beta = F(t_curr, u_curr)  # fixed second argument
        t_curr = t_curr + h

        # Initial guess: explicit Euler
        u_temp = u_curr + h * beta 
        u_guess = u_temp

        # Fixed-point iteration
        for _ in range(max_iter):
            alpha = F(t_curr, u_temp)
            u_guess = u_curr + h * A(alpha, beta, zero_tol=zero_tol)

            if abs(u_guess - u_temp) < tol:
                break

            u_temp = u_guess
        else:
            print(f"Warning: fixed-point iteration did not converge at step {k+1}")

        # Update for next step
        u_curr = u_guess    

        t_history.append(t_curr)
        u_history.append(u_curr)

def _of_Type_6(F, h, tol, zero_tol, max_iter, steps, t_curr, u_curr, t_history, u_history):
    """
    Implements the specular Euler scheme of Type 6.
    """
    for k in tqdm(range(steps), desc="Running the specular Euler scheme of Type 6"):
        t_next = t_curr + h

        # Initial guess: explicit Euler
        u_temp = u_curr + h * F(t_curr, u_curr)
        u_guess = u_temp

        # Fixed-point iteration
        for _ in range(max_iter):
            alpha = F(t_next, u_temp)
            beta = (u_temp - u_curr) / h

            u_guess = u_curr + h * A(alpha, beta, zero_tol=zero_tol)

            if abs(u_guess - u_temp) < tol:
                break
            
            u_temp = u_guess
        else:
            print(f"Warning: fixed-point iteration did not converge at step {k+1}")

        # Update for next step
        t_curr, u_curr = t_next, u_guess

        t_history.append(t_curr)
        u_history.append(u_curr)

def trigonometric_scheme(
    F: Callable[[float], float],
    t_0: float, 
    u_0: Callable[[float], float] | float,
    u_1: Callable[[float], float] | float,
    T: float, 
    h: float = 1e-6
) -> ODEResult:
    """
    Solves an initial value problem (IVP) using the specular trigonometric scheme.

    Parameters
    ----------
    F : callable
        The given source function ``F`` in (IVP).
        The calling signature should be ``F(t, u)`` where ``t`` and ``u`` are scalars.
    t_0 : float
        The starting time of the simulation.
    u_0 : callable
        The given initial condition ``u_0`` in (IVP).
    u_1 : callable | float | bool
        The numerical solution at the time ``t_1 = t_0 + h`` for Types 1, 2, and 3.
        If a float or callable is provided, it is used as the exact value.
        If False, the explicit Euler scheme is applied.
        Default: ``False``.
    T : float
        The end time of the simulation.
    h : float, optional
        Mesh size used in the finite difference approximation. Must be positive.
        Default: ``1e-6``.

    Returns
    -------
    ODEResult
        An object containing ``(t, u)`` data and the scheme name.
    """
    t_prev = t_0
    u_prev = u_0(t_0) if callable(u_0) else u_0

    t_curr = t_0 + h
    u_curr = u_1(t_curr) if callable(u_1) else u_1

    all_history = {}
    t_history = [t_prev, t_curr]
    u_history = [u_prev, u_curr]

    steps = int((T - t_0) / h)

    for m in tqdm(range(steps - 1), desc="Running specular trigonometric scheme"):
        t_next = t_curr + h
        u_next = u_curr + h*math.tan(2*math.atan(F(t_curr, u_curr)) - math.atan((u_curr - u_prev) / h)) # type: ignore

        t_history.append(t_next)
        u_history.append(u_next)

        t_prev, u_prev = t_curr, u_curr
        t_curr, u_curr = t_next, u_next

    all_history["variables"] = np.array(t_history)
    all_history["values"] = np.array(u_history)

    return ODEResult(
        scheme="specular trigonometric scheme",
        h=h,
        all_history=all_history
    )




import os
import numpy as np
from typing import Tuple
import matplotlib.pyplot as plt

class OptimizationResult:
    def __init__(
            self, 
            method: str,
            solution: np.ndarray, 
            func_val: int | float | np.number, 
            iteration: int, 
            runtime: float,
            all_history: dict
    ):
        self.method = method
        self.solution = solution
        self.func_val = func_val
        self.iteration = iteration
        self.runtime = runtime
        self.all_history = all_history

    def __repr__(self):
        return (
            f"[{self.method}]\n"
            f"    solution: {self.solution}\n"
            f"  func value: {self.func_val}\n"
            f"   iteration: {self.iteration}"
        )
    
    def last_record(
        self
    ) -> Tuple[float, float, float]:
        """
        Returns the final solution x, the value of f at x, and the runtime as a tuple.
        
        Returns
        -------
        Tuple[float, float]
            (x, f(x), runtime)
        """
        return self.solution, self.func_val, self.runtime # type: ignore

    def history(
        self
    ) -> Tuple[np.ndarray, np.ndarray, float]:
        """
        Returns the time grid and the numerical solution as a tuple.
        
        Returns
        -------
        Tuple[np.ndarray, np.ndarray, float]
            (x_history, f_history, runtime)
        """
        return self.all_history["variables"], self.all_history["values"], self.runtime


from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

long_description = long_description.replace(
    "docs/figures/",
    "https://raw.githubusercontent.com/kyjung2357/specular-differentiation/main/docs/figures/"
)

long_description = long_description.replace(
    "docs/README.md",
    "https://github.com/kyjung2357/specular-differentiation/blob/main/docs/README.md"
)

setup(
    long_description=long_description,
    long_description_content_type="text/markdown",
)


import numpy as np
from tqdm import tqdm
import time
import inspect
from typing import Callable, TypeAlias, Sequence
from .result import OptimizationResult
from .step_size import StepSize
from ..calculation import derivative, gradient

SUPPORTED_METHODS = ['specular gradient', 'implicit', 'stochastic', 'hybrid']

ComponentFunc: TypeAlias = Callable[[int | float | np.number | list | np.ndarray], int | float | np.number]

def gradient_method(
    f: Callable[[int | float | np.number | list | np.ndarray], int | float | np.number],
    x_0: int | float | list | np.ndarray,
    step_size: StepSize,
    h: float = 1e-6,
    form: str = 'specular gradient',
    tol: float = 1e-6,
    zero_tol: float = 1e-8,
    max_iter: int = 1000,
    f_j: Sequence[ComponentFunc] | Callable | None = None,
    m: int = 1,
    switch_iter: int | None = 2,
    record_history: bool = True,
    print_bar: bool = True
) -> OptimizationResult:
    """
    The specular gradient method for minimizing a nonsmooth convex function.

    Parameters
    ----------
    f : callable
        The objective function to minimize.
    x_0 : int | float | list | np.ndarray
        The starting point for the optimization.
    step_size : StepSize
        The step size `h_k`.
    h : float, optional
        Mesh size used in the finite difference approximation. Must be positive.
        Default: ``1e-6``.
    form : str, optional
        The form of the specular gradient method.
        Supported forms: ``'specular gradient'``, ``'implicit'``, ``'stochastic'``, ``'hybrid'``.
        Default: ``'specular gradient'``.
    tol : float, optional
        Tolerance for iterations.
        Default: ``1e-6``.
    zero_tol : float, optional
        A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.
        Default: ``1e-8``.
    max_iter : int, optional
        Maximum number of iterations.
        Default: ``1000``.
    f_j : sequence of callable | callable | None, optional
        The component function of ``f``.
        Used for the stochastic and hybrid forms to compute a random component of the objective function.
        Default: ``None``.

        * If a sequence of callables is provided, each callable should accept a single argument (the variable `x`).

        * If a single callable is provided, it should accept two arguments: the variable `x` and an index `j`, and return the `j`-th component function value at `x`.
    m : int, optional
        The number of component functions.
        Used for the stochastic and hybrid forms.
    switch_iter : int | None, optional
        The iteration to switch from a method to another for the hybrid form.
        Used for the hybrid form only.
        Default: ``2``.
    record_history : bool, optional
        Whether to record the history of variables and function values.
        Default: ``True``.
    print_bar : bool, optional
        Whether to print the progress bar.
        Default: ``True``.

    Returns
    -------
    OptimizationResult
        The result of the optimization containing the solution, function value, number of iterations, runtime, and history.
    
    Raises
    ------
    ValueError
        If ``h`` is not positive.
    TypeError
        If an unknown ``form`` is provided.
    """

    if h is None or h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")
    
    x = np.array(x_0, dtype=float).copy()
    n = x.size
    
    all_history = {}
    x_history = []
    f_history = []

    start_time = time.time()

    # the n-dimensional case
    if n > 1:
        if form == 'specular gradient':
            res_x, res_f, res_k = _vector(f, f_history, x, x_history, step_size, h, tol, zero_tol, max_iter, record_history, print_bar)

        elif form == 'stochastic':
            if f_j is None:
                raise ValueError("Component functions 'f_j' must be provided for the stochastic form.")

            form = 'stochastic specular gradient'
            res_x, res_f, res_k = _vector_stochastic(f, f_history, x, x_history, step_size, h, tol, zero_tol, f_j, m, max_iter, record_history, print_bar) # type: ignore

        elif form == 'hybrid':
            if f_j is None:
                raise ValueError("Component functions 'f_j' must be provided for the stochastic form.")
            
            form = 'hybrid specular gradient'
            switch_iter = switch_iter if switch_iter is not None else max_iter
            remaining_iter = max_iter - switch_iter

            res_x, res_f, res_k = _vector(f, f_history, x, x_history, step_size, h, tol, zero_tol, switch_iter, record_history, print_bar)
            res_x, res_f, res_k = _vector_stochastic(f, f_history, res_x, x_history, step_size, h, tol, zero_tol, f_j, m, remaining_iter, record_history, print_bar) # type: ignore

        else:
            raise TypeError(f"Unknown form '{form}'. Supported forms: {SUPPORTED_METHODS}")

    # the one-dimensional case
    elif n == 1:
        x = x.item()

        if form == 'specular gradient':
            res_x, res_f, res_k = _scalar(f, f_history, x, x_history, step_size, h, tol, zero_tol, max_iter, record_history, print_bar)
            
        elif form == 'implicit':
            form = 'implicit specular gradient'
            res_x, res_f, res_k = _scalar_implicit(f, f_history, x, x_history, step_size, h, tol, max_iter, record_history, print_bar)
            
        else:
            raise TypeError(f"Unknown form '{form}'. Supported forms: {SUPPORTED_METHODS}")
    
    else:
        raise TypeError(f"Unknown form '{form}'. Supported forms: {SUPPORTED_METHODS}")
    
    runtime = time.time() - start_time

    if record_history:
        all_history["variables"] = x_history
        all_history["values"] = f_history

    return OptimizationResult(
        method=form,
        solution=res_x,
        func_val=res_f,
        iteration=res_k,
        runtime=runtime,
        all_history=all_history
    ) 

def _scalar(
    f: Callable[[int | float | np.number], int | float | np.number | list | np.ndarray],
    f_history: list,
    x: int | float,
    x_history: list,
    step_size: StepSize,
    h: float,
    tol: float,
    zero_tol: float,
    max_iter: int,
    record_history: bool,
    print_bar: bool
) -> tuple:
    """
    Scalar implementation of :func:`gradient_method`.
    The specular gradient method in the one-dimensional case.
    """
    k = 1

    for _ in tqdm(range(1, max_iter + 1), desc="Running the specular gradient method", disable=not print_bar, leave=False):
        if record_history is True:
            x_history.append(x)
            f_history.append(f(x))

        specular_derivative = derivative(f=f, x=x, h=h, zero_tol=zero_tol)
        norm = np.linalg.norm(specular_derivative)
        if norm < tol:
            break
        
        x -= step_size(k)*(specular_derivative / norm) # type: ignore
        k += 1
    
    return x, f(x), k

def _scalar_implicit(
    f: Callable[[int | float | np.number], int | float | np.number],
    f_history: list,
    x: int | float,
    x_history: list,
    step_size: StepSize,
    h: float,
    tol: float,
    max_iter: int,
    record_history: bool,
    print_bar: bool
) -> tuple:
    """
    Scalar implementation of :func:`gradient_method`.
    The implicit specular gradient method in the one-dimensional case.
    """
    k = 1

    for _ in tqdm(range(1, max_iter + 1), desc="Running the implicit specular gradient method", disable=not print_bar, leave=False):
        if record_history is True:
            x_history.append(x)
            f_history.append(f(x))

        sum_of_one_sided_derivatives = (f(x + h) - f(x - h)) / h

        if abs(sum_of_one_sided_derivatives) < tol:
            break
        
        x -= step_size(k)*(sum_of_one_sided_derivatives / abs(sum_of_one_sided_derivatives))
        k += 1
    
    return x, f(x), k

def _vector(
    f: Callable[[list | np.ndarray], int | float | np.number],
    f_history: list,
    x: list | np.ndarray,
    x_history: list,
    step_size: StepSize,
    h: float,
    tol: float,
    zero_tol: float,
    max_iter: int, 
    record_history: bool,
    print_bar: bool
) -> tuple:
    """
    Vector implementation of :func:`gradient_method`.
    The specular gradient method in the n-dimensional case.
    """
    k = 1

    for _ in tqdm(range(1, max_iter + 1), desc="Running the specular gradient method", disable=not print_bar, leave=False):
        if record_history is True:
            x_history.append(x)
            f_history.append(f(x))

        specular_gradient = gradient(f=f, x=x, h=h, zero_tol=zero_tol)
        norm = np.linalg.norm(specular_gradient)

        if norm < tol:
            break

        x -= step_size(k)*(specular_gradient / norm)
        k += 1
    
    return x, f(x), k

def _vector_stochastic(
    f: Callable[[list | np.ndarray], int | float | np.number],
    f_history: list,
    x: list | np.ndarray,
    x_history: list,
    step_size: StepSize,
    h: float,
    tol: float,
    zero_tol: float,
    f_j: Sequence[ComponentFunc],
    m: int = 1,
    max_iter: int = 1000, 
    record_history: bool = True,
    print_bar: bool = True
) -> tuple:
    """
    Vector implementation of :func:`gradient_method`.
    The stochastic specular gradient method in the n-dimensional case.
    """
    k = 1

    for _ in tqdm(range(1, max_iter + 1), desc="Running the stochastic specular gradient method", disable=not print_bar, leave=False):
        if record_history is True:
            x_history.append(x)
            f_history.append(f(x)) 

        if hasattr(f_j, '__len__'):
            num_components = len(f_j)
        else:
            num_components = m
        
        # A random index j is selected at each iteration
        j = np.random.randint(num_components)

        try:
            component_func = f_j[j]

        except (TypeError, IndexError):
            if not callable(f_j):
                raise TypeError(f"f_j must be a list of functions or a callable. Got {type(f_j)} instead.")
            
            sig = inspect.signature(f_j)
            params = list(sig.parameters.values())

            has_varargs = any(p.kind in (p.VAR_POSITIONAL, p.VAR_KEYWORD) for p in params)

            if len(params) < 2 and not has_varargs:
                raise ValueError(
                    f"The function f_j must accept at least 2 arguments (x and index). "
                    f"Current signature is: {sig}"
                )

            component_func = lambda x_val: f_j(x_val, j)

        component_specular_gradient = gradient(f=component_func, x=x, h=h, zero_tol=zero_tol)
        norm = np.linalg.norm(component_specular_gradient)

        if norm < tol:
            break

        x -= step_size(k)*(component_specular_gradient / norm)
        k += 1

    return x, f(x), k